#
#  Copyright 1996-2014 Jim Pirzyk / pirzyk.org
#  All rights reserved.
#
#  Security::OS::linux - Linux specific configuration information
#
package Security::OS::linux;

use Security::util::detaint;
use Security::util::die;
use Security::util::print;
use Security::util::read_config;
use Security::util::stat;
use Security::util::users;
use Security::util::groups;
use Security::util::uniq;
use Security::util::MD5;
use Exporter;
use Fcntl qw(:mode);
use strict;
use warnings "all";

our @ISA = qw(Exporter);
our @EXPORT_OK = qw( init check );

our (%perms, %perms2, %files, %pkgs, %sigs);
our $rpm = '/bin/rpm';
our $chkstat = '/usr/bin/chkstat';
our $aptkey = '/usr/bin/apt-key list';
our $debsums = '/usr/bin/debsums -s';

sub init {
	print ( STDERR  "In Security::OS::linux::init()\n") if ( $main::opts{'d'} );

	&read_config('ignorepkgfiles');
	&read_config('nopkgduperr');
	&read_config('gpgkeys');

	$main::showmount = '/usr/sbin/showmount';
	$main::exportsfile = '/etc/exports';
	$main::securefile = '/etc/securetty';
	$main::kerneldir = '/lib/modules';
	$main::kernelmodlist = '/usr/bin/awk \'{print $1}\' /proc/modules';
	$main::sudofile = '/etc/sudoers';
}

sub check {
	print ( STDERR  "In Security::OS::linux::check()\n") if ( $main::opts{'d'} );

	&Security::rootlogin::check_securettys() 
		if ( defined &Security::rootlogin::check_securettys );

	# Do we have an RPM based linux distro?
	if ( -f $rpm ) {
		&check_rpm_files();

		# Check for missing or untrusted package signers
		&check_rpm_signers();
	}

	&check_dpkg_signers()
		if ( -f '/etc/apt/trusted.gpg' );

	&check_dpkg_files()
		if ( -f '/usr/bin/debsums' );

	&check_permissions() if ( -f $chkstat );
}

sub check_dpkg_files {
	my ($FP);
	print ( STDERR  "In Security::OS::linux::check_dpkg_files()\n") if ( $main::opts{'d'} );

	open ($FP, "$debsums 2>&1 |") || &security_die("Could not run $debsums command\n");
	while (<$FP>) {
		chomp;

		# Should we really complain about this?
		if ( /debsums: no md5sums for (\S+)/ ) {
			$perms{$1} = "package has no md5sums"
				if ( !grep (/^\Q${1}\E$/, @{$main::config{'ignorepkgfiles'}}));
		}

		if ( /debsums: missing file (\S+) \(from (\S+) package\)/) {
			$perms{$1} = "file is missing (package $2)"
				if ( !grep (/^\Q${1}\E$/, @{$main::config{'ignorepkgfiles'}}));
		}

		if ( /debsums: changed file (\S+) \(from (\S+) package\)/) {
			$perms{$1} = "checksum mismatch (package $2)"
				if ( !grep (/^\Q${1}\E$/, @{$main::config{'ignorepkgfiles'}}) && ${1} !~ m|^/usr/share/doc/.*/changelog\.Debian\.gz$| );
		}
	}
	close $FP;
}

sub check_rpm_files {
	print ( STDERR  "In Security::OS::linux::check_rpm_files()\n") if ( $main::opts{'d'} );

	_get_rpm_errors();
	_find_rpm_info();

	# Now go through and tell us what is wrong....
	_save_rpm_errors();
}

sub check_dpkg_signers {
	my ($FP, @found);
	print ( STDERR  "In Security::OS::linux::check_dpkg_signers()\n") if ( $main::opts{'d'} );

	return if ( ! scalar @{$main::config{'gpgkeys'}} );

	open($FP, "$aptkey |") || &security_die("Could not run $aptkey command\n");
	while (<$FP>) {
		chomp;

		if ( /^pub / ) {
			my ($pub, $keysize, $key, $date) = &detaint('apt-key', split (m|[/\s\t]+|, $_));
			$sigs{$key} = 'untrusted key loaded into GPG keyring'
				if ( !grep (/^\Q${key}\E$/, @{$main::config{'gpgkeys'}}));
		}
	}
	close $FP;
}

sub check_rpm_signers {
	my ($FP, @found);
	print ( STDERR  "In Security::OS::linux::check_rpm_signers()\n") if ( $main::opts{'d'} );

	# Don't do anything if we don't have a list of valid GPG Keys
	return if ( ! scalar @{$main::config{'gpgkeys'}} );

	open($FP, "$rpm -qa --qf \"%{NAME}-%{VERSION}-%{RELEASE} %{SIGPGP:pgpsig}\n\" 2>&1 |") || &security_die("Could not run $rpm command\n");
	while (<$FP>) {
		chomp;
		my ($pkg, @sig) = &detaint('pgp sigs', split());

		# Now try the GPG signature instead of PGP
		if ( !scalar @sig || $sig[0] eq '(none)' ) {
			my $res=`$rpm -q --qf \"%{NAME}-%{VERSION}-%{RELEASE} %{SIGGPG:pgpsig}\n\" $pkg 2>&1`;
			chomp $res;
			($pkg, @sig) = &detaint('gpg sigs', split (/[\s\t]+/, $res));
		}

		# failed to find a signature
		if ( ! scalar @sig || $sig[0] eq '(none)' ) {

			# Complain about no signer, except for the trusted gpg-pubkey themselves.
			if ( $pkg !~ /^gpg-pubkey-/ ) {
				$sigs{$pkg} = 'no signer'
			} else {
				my ($pkg2) = $pkg;
				$pkg2 =~ s/gpg-pubkey-//;
				$pkg2 =~ s/-[^-]*$//;
				if ( !grep (/^\Q${pkg2}\E$/, @{$main::config{'gpgkeys'}})) {
					$sigs{$pkg} = 'untrusted key loaded into RPM database';
				} else {
					push @found, $pkg2;
				}
			}
		} elsif ( $sig[-3] eq 'Key' && $sig[-2] eq 'ID' ) {
			my @a = split(//, $sig[-1]);
			my $key = join('', @a[8..15]);
			$sigs{$pkg} = 'RPM signed by untrusted key'
				if ( !grep (/^\Q${key}\E$/, @{$main::config{'gpgkeys'}}));
		} else {
			$sigs{$pkg} = 'could not find signer key';
		}
	}
	close $FP;

	foreach my $key (@{$main::config{'gpgkeys'}}) {
		$sigs{"gpg-pubkey-$key"} = 'trusted key not installed'
			if ( !grep (/^\Q${key}\E$/, @found));
	}
}

sub check_permissions {
	my (@files) = ( '/etc/permissions' ) if ( -f '/etc/permissions' );
	print ( STDERR  "In Security::OS::linux::check_permissions()\n") if ( $main::opts{'d'} );
	my ($policy, @policies);

	# First determine the policies we need to use.
	if ( -f '/etc/sysconfig/security' ) {
		my $FP;
		open ($FP, '/etc/sysconfig/security') || &security_die("Could not read /etc/sysconfig/security\n");
		grep (/^PERMISSION_SECURITY="([^"]+)"/ && ($policy=$1), <$FP>);
		close $FP;
		@policies = &detaint('policies', split(/\s+/, $policy));
	}

	# Now find all the policy files and add them.
	for my $policy ( @policies ) {
		push @files, "/etc/permissions.$policy" if ( -f "/etc/permissions.$policy" );
		my $DIR;
		if ( opendir ($DIR, '/etc/permissions.d')) {  
			while (my $file = &detaint('readdir', readdir ($DIR))) {
				my $f2 = "/etc/permissions.d/$file";
				if ( -f $f2 ) {
					# Add the non-policy extended files first (if not already there).
					push @files, $f2 if ( $file !~ /\S+\.\S+/ && ! grep (/^\Q${f2}\E$/, @files) );
					# Now add the policy specific ones if we have a policy match.
					push @files, $f2 if ( $file =~ /\.$policy$/ );
				}
			}
			closedir $DIR;
		}
	}

	# Now run the command and save any ouput we were told not to ignore.
	my $cmd = $chkstat . ' ' . join (' ', @files);
	foreach my $line (&detaint('chkstat', split (/\n/, `$cmd`))) {
		if ( defined $line && $line =~ /^(\S+) should be / ) {
			if ( !grep (/^\Q${1}\E$/, @{$main::config{'ignorepkgfiles'}})) {
				$perms2{$1} = $line;
				$perms2{$1} =~ s/^$1 //;
			}
		}
	}
}

sub print {
	my ($str);
	print ( STDERR  "In Security::OS::linux::print()\n") if ( $main::opts{'d'} );

	# Flatten out the perl stucture for printing...
	for my $file ( keys %perms ) {
		if ( ref $perms{$file} eq 'HASH' ) {
			if ( scalar keys %{$perms{$file}} ) {
				for my $type ( keys %{$perms{$file}} ) {
					if ( exists $perms{$file}->{$type}->{'is'} ) { 
						$str = _format_line($type, $perms{$file}->{$type}->{'is'}, &Security::util::uniq::uniq(@{$perms{$file}->{$type}->{'should'}}));
						delete $perms{$file}->{$type}->{'is'};
					} else {
						$str = _format_line($type, 'N/A', &Security::util::uniq::uniq(@{$perms{$file}->{$type}->{'should'}}));
					}
					delete $perms{$file}->{$type}->{'should'};
					$perms{$file}->{$type} = $str;
				}
			} else {
				delete $perms{$file};
			}
		}
	}

	&print_data ("\n\nchecking installed package integrity:\n", %perms);
	&print_data ("\n\nchecking package signatures:\n", %sigs);
	&print_data ("\n\nchecking chkstat command output:\n", %perms2);
}

sub _get_rpm_errors {
	my ($FP);
	open ($FP, "$rpm --verify --all 2>&1 |") || &security_die ("Could not run $rpm --verify --all\n");
	while (<$FP>) {
		chomp;
		s/ \(Permission denied\)$//;
		my ($fmode, $type, $file) = split;
		$file = $type if ( ! defined $file );

		if ( -e $file || $file =~ m|^/| ) {

			# If we have a config file and have modified the (S)ize,
			# MD(5) Checksum or m(T)ime, ignore the change.
			if ( !grep (/^\Q${file}\E$/, @{$main::config{'ignorepkgfiles'}}) && !($type eq 'c' && ( $fmode =~ /S/ || $fmode =~ /5/ || $fmode =~ /T/ ))) {
				if ( $fmode ne 'missing' ) {
					my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = &Security::util::stat::get_stat($file, 'lstat');
					my $owner = &Security::util::users::get('uid', $uid, 'name');
					my $group = &Security::util::groups::get('gid', $gid, 'name');
					$files{$file} = {
						size => $size,
						mode => $mode,
						owner => (defined $owner? $owner: $uid),
						group => (defined $group? $group: $gid),
						rdev => $rdev,
						mtime => $mtime,
						failure => exists $files{$file}->{'failure'} ? $files{$file}->{'failure'} . $fmode : $fmode,
					};

					# Only generate the MD5 checksum if we have a file.
					if ( &S_ISREG($mode) ) {
						$files{$file}->{'md5'} = &Security::util::MD5::check($file);
						chomp $files{$file}->{'md5'};
						$files{$file}->{'md5'} =~ s/MD5 \($file\) = //;
					}
				} else {
					$files{$file}->{'failure'} = $fmode;
				}
			}
		} elsif ( $fmode eq 'prelink:' ) {
			$file = $type;
			$file =~ s/:$//;
			$files{$file}->{'failure'} = exists $files{$file}->{'failure'} ? $files{$file}->{'failure'} . 'K' : 'K';
		} else {
			warn "Could not understand output ($_)\n";
		}
	}
	close $FP;
}

sub _find_rpm_info {
	foreach my $file ( keys %files ) {
		foreach my $pkg (&detaint('pkg', split (/\n/, `$rpm -qf --qf "%{NAME}\n" $file 2>&1`))) {
			push @{$files{$file}->{pkg}}, $pkg;
			if ( ! exists $pkgs{$pkg} ) {
				my $FP;
				open ($FP, "$rpm -q --dump $pkg 2>&1 |") || &security_die ("Could not run $rpm -q --dump $pkg\n");
				while (<$FP>) {
					chomp;
					my ($file, $size, $mtime, $md5, $mode, $owner, $group, $isconfig, $isdoc, $rdev, $symlink) = &detaint('dump', split());
					$pkgs{$pkg}->{$file} = {
						size => $size,
						mtime => $mtime,
						md5 => $md5,
						mode => oct($mode),
						owner => $owner,
						group => $group,
						rdev => $rdev,
						symlink => $symlink,
					}
				}
				close $FP;
			}
		}
	}
}

sub _save_rpm_errors {
	foreach my $file ( keys %files ) {
		foreach my $pkg ( @{$files{$file}->{'pkg'}} ) {
			if ( exists $pkgs{$pkg} ) {
				if ( $files{$file}->{'failure'} eq 'missing' ) {
					push @{$perms{$file}->{'file type'}->{'should'}}, &_format_type($pkgs{$pkg}->{$file}->{'mode'});
					push @{$perms{$file}->{'mode'}->{'should'}}, &_format_mode($pkgs{$pkg}->{$file}->{'mode'});
					push @{$perms{$file}->{'device'}->{'should'}}, $pkgs{$pkg}->{$file}->{'rdev'}
						if ( $pkgs{$pkg}->{$file}->{'rdev'} > 0 );
					push @{$perms{$file}->{'link'}->{'should'}}, $pkgs{$pkg}->{$file}->{symlink}
						if (&S_ISLNK($pkgs{$pkg}->{$file}->{mode}));

					push @{$perms{$file}->{'group'}->{'should'}}, $pkgs{$pkg}->{$file}->{'group'};
					push @{$perms{$file}->{'owner'}->{'should'}}, $pkgs{$pkg}->{$file}->{'owner'};
				} else {
					if ( $files{$file}->{'failure'} =~ /S/ ) {
						$perms{$file}->{'size'}->{'is'} = $files{$file}->{'size'};
						push @{$perms{$file}->{'size'}->{'should'}}, $pkgs{$pkg}->{$file}->{'size'};
						if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'size'} eq $pkgs{$pkg}->{$file}->{'size'} ) {
							$files{$file}->{'failure'} =~ s/S/\./;
							delete $perms{$file}->{'size'};
						}
					}

					if ( $files{$file}->{'failure'} =~ /M/ ) {
						if (&S_IFMT($pkgs{$pkg}->{$file}->{mode}) != &S_IFMT($files{$file}->{mode})) {
							$perms{$file}->{'file type'}->{'is'} = &_format_type($files{$file}->{'mode'});
							push @{$perms{$file}->{'file type'}->{'should'}}, &_format_type($pkgs{$pkg}->{$file}->{'mode'});
							if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'mode'} eq $pkgs{$pkg}->{$file}->{'mode'} ) {
								$files{$file}->{'failure'} =~ s/M/\./;
								delete $perms{$file}->{'file type'};
							}
						} else {
							$perms{$file}->{'mode'}->{'is'} = &_format_mode($files{$file}->{'mode'});
							push @{$perms{$file}->{'mode'}->{'should'}}, &_format_mode($pkgs{$pkg}->{$file}->{'mode'});
							if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'mode'} eq $pkgs{$pkg}->{$file}->{'mode'} ) {
								$files{$file}->{'failure'} =~ s/M/\./;
								delete $perms{$file}->{'mode'};
							}
						}
					}
					
					if ( $files{$file}->{'failure'} =~ /5/ ) {
						$perms{$file}->{'md5'}->{'is'} = $files{$file}->{'md5'};
						push @{$perms{$file}->{'md5'}->{'should'}}, $pkgs{$pkg}->{$file}->{'md5'};
						if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'md5'} eq $pkgs{$pkg}->{$file}->{'md5'} ) {
							$files{$file}->{'failure'} =~ s/5/\./;
							delete $perms{$file}->{'md5'};
						}
					}

					if ( $files{$file}->{'failure'} =~ /D/ ) {
						$perms{$file}->{'device'}->{'is'} = $files{$file}->{'rdev'};
						push @{$perms{$file}->{'device'}->{'should'}}, $pkgs{$pkg}->{$file}->{'rdev'};
						if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'rdev'} eq $pkgs{$pkg}->{$file}->{'rdev'} ) {
							$files{$file}->{'failure'} =~ s/D/\./;
							delete $perms{$file}->{'device'};
						}
					}

					if ( $files{$file}->{'failure'} =~ /L/ ) {
						if ( $pkgs{$pkg}->{$file}->{symlink} eq 'X' ) {
							push @{$perms{$file}->{'link'}->{'should'}}, 'should not be a link';
						} else {
							push @{$perms{$file}->{'link'}->{'should'}}, $pkgs{$pkg}->{$file}->{symlink};
						}
					}

					if ( $files{$file}->{'failure'} =~ /U/ ) {
						$perms{$file}->{'owner'}->{'is'} = $files{$file}->{'owner'};
						push @{$perms{$file}->{'owner'}->{'should'}}, $pkgs{$pkg}->{$file}->{'owner'};
						if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'owner'} eq $pkgs{$pkg}->{$file}->{'owner'} ) {
							$files{$file}->{'failure'} =~ s/U/\./;
							delete $perms{$file}->{'owner'};
						}
					}

					if ( $files{$file}->{'failure'} =~ /G/ ) {
						$perms{$file}->{'group'}->{'is'} = $files{$file}->{'group'};
						push @{$perms{$file}->{'group'}->{'should'}}, $pkgs{$pkg}->{$file}->{'group'};
						if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'group'} eq $pkgs{$pkg}->{$file}->{'group'} ) {
							$files{$file}->{'failure'} =~ s/G/\./;
							delete $perms{$file}->{'group'};
						}
					}

					if ( $files{$file}->{'failure'} =~ /T/ ) {
						$perms{$file}->{'mtime'}->{'is'} = $files{$file}->{'mtime'};
						push @{$perms{$file}->{'mtime'}->{'should'}}, $pkgs{$pkg}->{$file}->{'mtime'};
						if ( $main::config{'nopkgduperr'} eq 'yes' && $files{$file}->{'mtime'} eq $pkgs{$pkg}->{$file}->{'mtime'} ) {
							$files{$file}->{'failure'} =~ s/T/\./;
							delete $perms{$file}->{'mtime'};
						}
					}

					# The prelink case
					if ( $files{$file}->{'failure'} =~ /K/ ) {
						push @{$perms{$file}->{'prelink'}->{'should'}}, 'needs to be prelinked';
					}

					# Not sure what to do for this case...
					if ( $files{$file}->{'failure'} =~ /P/ ) {
					}
				}
			} else {
				&security_die ("Can't find package for file ($file)\n");
			}
		}
	}
}

sub _format_line {
	my ($type, $is, @should) = @_;
	my ($str);

	if ( $type eq 'link' ) {
		if ( grep /should not be a link/, @should ) {
			$str = 'should not be a link';
		} else {
			if ( scalar @should gt 1 ) {
				$str = ' -> {' . join ('|', sort(@should)) . '}';
			} else {
				$str = '-> ' . $should[0];
			}
		}
	} else {
		if ( grep /needs to be prelinked/, @should ) {
			$str = 'needs to be prelinked';
		} elsif ( scalar @should gt 1 ) {
			$str = '({' . join ('|', sort(@should)) . '}:' . $is . ')';
		} else {
			$str = '(' . $should[0] . ':' . $is . ')';
		}
	}

	return $str;
}

sub _format_type {
	my ($mode) = @_;
	return 'file' if (&S_ISREG($mode));
	return 'directory' if (&S_ISDIR($mode));
	return 'link' if (&S_ISLNK($mode));
	return 'socket' if (&S_ISSOCK($mode));
	return 'block' if (&S_ISBLK($mode));
	return 'character' if (&S_ISCHR($mode));
	return 'fifo' if (&S_ISFIFO($mode));
}

sub _format_mode {
	my ($mode) = @_;
	return sprintf "%4.4o", &S_IMODE($mode);
}

1;

__END__

=head1 NAME

Security::OS::linux

=head1 SYNOPSIS

Security::OS::linux::init();

Security::OS::linux::check();

=head1 DESCRIPTION

Sets variables for OS specific commands.

=head1 API

Defines the following API calls

=over 4

=item C<init()>

Init sets the following global variables:

=over 4

=item C<$main::showmount>

The location of the showmount command, used in the C<Security::exports> module.

=item C<$main::exportsfile>

The location of the exports file, used in C<Security::exports> module.

=item C<$main::sudofile>

The location of F<sudoers> file, used in C<Security::sudo> module.

=item C<$main::kerneldir>

The location of the kernel loadable modules, used in C<Security::kernel>
module.

=item C<$main::securefile>

The location of the list of ports root is allowed in on, used in
C<Security::rootlogin> module.

=back

=item C<check()>

Calls C<Security::rootlogin::check_securetty()> to check what root ports are
configured and compares it to what ports we know about.

If the L<rpm> command is found on the system, it will run L<rpm -q --verify...>
to find inconsitancies between the RPM database and what is on the filesystem.

The program will also check to find and report any RPM packages that are not
signed or are not signed by a key listed in C<gpgkeys>.  It will report any
keys that are installed but not listed in C<gpgkeys> as well as reporting
any keys in C<gpgkeys> not installed.

If the L<chkstat> command is found on the system (OpenSuSE, etc), it will run
L<chkstat /etc/permissions ...> to find inconsitancies between the permission
files and what is on the filesystem.

If the L<debsums> command is found on the system (Ubuntu/Debian), it will run
this command to find any checksum mismatches or missing files.  Use the 
C<ignorepkgfiles> config option to have the check ignore packages that are not
shipped with md5 checksums.

If the L<apt-key> command is found and C<gpgkeys> is set, it will verify that
all the keys that it reports on the system are listed in C<gpgkeys>.

=back

=head1 CONFIGURATION OPTIONS

The script will call C<read_config> on the following config file options:

=over 4

=item C<ignorepkgfiles>

List of files to ignore errors returned from L<rpm --verify> and L<chkstat>.

=item C<nopkgduperr>

This option will skip package errors if the problem file has multiple
entries that do not match one of the entries.  For example if a directory
should be owned by C<root> in one package and by C<bin> in another and
the directory is owned by C<root>.

=item C<gpgkeys>

This option lists all the valid gpg keys that are allowed to be installed
on the system and used to sign installed packages.

=back

=head1 HISTORY

Checking RPM package signers and Debian checksums were proposed on
L<http://article.gmane.org/gmane.linux.kernel/1197924>

=head1 SEE ALSO

L<Security::util::groups>

L<Security::util::stat>

L<Security::util::uniq>

L<Security::util::users>

L<apt-key(8)>

L<chkstat(8)>

L<debsums(8)>

L<rpm(8)>

=cut
