#
#  Copyright 1996-2014 Jim Pirzyk / pirzyk.org
#  All rights reserved.
#
#  Security::library - Shared library control module
#
package Security::library;

use Security::util::print;
use Security::find;
use Security::util::stat;
use Security::db;
use Exporter;
use POSIX;
use strict;
use warnings "all";

our @ISA = qw(Exporter);
our @EXPORT_OK = qw( init print );

our (%libraryperms, $sharedregex);

sub init {
	print ( STDERR  "In Security::library::init()\n") if ( $main::opts{'d'} );

	&Security::find::add_wanted (\&library_wanted);

	$sharedregex = qr|^lib[^/]*\.so(\..*)?$|
		if ( ! defined $sharedregex );
}

# Added shared library file checks based on the SANS98 Conference
sub library_wanted {
	my ($file) = @_;
	if ( ! -l $file && -f $file ) {
		if ( $file =~ /$sharedregex/ ) {

			my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$curbytes) = 
				&Security::util::stat::get_stat($File::Find::name, 'lstat');

			&Security::db::verify($File::Find::name);

			# If the file is world writeable, then flag this case too
			$libraryperms{$File::Find::name} = 1
				if ( defined $mode && $mode & &S_IWOTH() );
		}
	}
}

sub print {
	print ( STDERR  "In Security::library::print()\n") if ( $main::opts{'d'} );

	&print_data ("\n\nchecking for world writable shared libraries:\n", %libraryperms);
}

1;

__END__

=head1 NAME

Security::library - shared library file check

=head1 SYNOPSIS

Security::library::init()

Security::library::print()

=head1 DESCRIPTION

Checks the checksum and file mode of all shared libraries we find on the
system.  The shared library is determined via C<sharedregex>.

=head1 API

This module implements the default API:

=over 4

=item C<init()>

Calls C<add_wanted (\&library_wanted)> to implement the match of files
that should be considered a shared library.

=item C<print()>

Calls C<print_data ()> to print out the list of shared libraries that were
world writable.

=back

=head1 CONFIGURATION OPTIONS

None

=head1 HISTORY

The idea for this came from the SANS 1998 Security conference in Orlando, FL.

=head1 SEE ALSO

L<Security::find>

L<Security::db>

L<Security::util::stat>

L<perlre(1)>

=cut

