#
#  Copyright 1996-2014 Jim Pirzyk / pirzyk.org
#  All rights reserved.
#
#  Security::rootlogin - root login control module
#
package Security::rootlogin;

use Security::util::die;
use Security::util::print;
use Security::util::read_config;
use Exporter;
use strict;
use warnings "all";
no warnings 'once';

our @ISA = qw(Exporter);
our @EXPORT_OK = qw( init check_ttys check_access check_defaultlogin print );

our (%ttys, $rootlogin, $umask);

sub init {
	print ( STDERR "In Security::rootlogin::init()\n") if ( $main::opts{'d'} );

	# Read the config file for root port definitions.
	&read_config('ignorerootports');

	$rootlogin = -1;
}

sub check_securettys {
	my $FP;

	print ( STDERR "In Security::rootlogin::check_securettys()\n") if ( $main::opts{'d'} );

	open ($FP, $main::securefile) || 
		&security_die ("Could not open login file $main::securefile\n");
	while (<$FP>) {
		chomp;
		if ( -c "/dev/$_" ) {
			$ttys{$_}=1;
		}
	}
	close $FP;

	&_remove_defaultports();
}

sub check_defaultlogin {
	my $FP;

	print ( STDERR "In Security::rootlogin::check_defaultlogin()\n") if ( $main::opts{'d'} );
	open ($FP, $main::loginfile) || &security_die ("Could not open login file $main::loginfile\n");
	$rootlogin=grep (/^CONSOLE/o, <$FP>);
	close $FP;

	&_remove_defaultports();
}

sub check_ttys {
	my $FP;

	print STDERR "In Security::rootlogin::check_ttys()\n"
		if ( $main::opts{'d'} );
	open ($FP, $main::securefile) || 
		&security_die ("Could not open login file $main::securefile\n");
	grep (/^([^ \t#]+)[ \t]*.*[ \t]*on[ \t]*secure/o && ($ttys{$1}=1), <$FP>);
	close $FP;

	&_remove_defaultports();
}

sub check_access {
	my $FP;

	print ( STDERR "In Security::rootlogin::check_access()\n") if ( $main::opts{'d'} );
	open ($FP, $main::accessfile) || 
		&security_die ("Could not open login file $main::accessfile\n");
	while (<$FP>) {
		chomp;
		my (@a) = split (/:/o, $_) if ( $_ !~ /^#/o );
		if ( scalar @a > 1 && $a[1] eq "root" ) {
			if ( $a[0] eq '+' ) {
				if ( $a[2] =~ /ALL$/o ) {
					$rootlogin=0;
				} else {
					my (@b) = split (/[ \t]+/o, $a[2]);
					for (my ($i) = 0; $i < scalar @b; $i++) {
						$ttys{$b[$i]} = 1 if ( $b[$i] ne 'ALL' );
					}
				}
			} else {
				if ( $a[2] =~ /ALL$/o ) {
					$rootlogin=1;
				} else {
					my (@b) = split (/[ \t]+/o, $a[2]);
					for (my ($i) = 0; $i < $#b; $i++) {
						$ttys{$b[$i]} = 1 
							if ( $b[$i] ne 'ALL' && 
								$b[$i] ne 'EXCEPT' );
					}
				}
			}
		}
	}
	close $FP;

	&_remove_defaultports();
}

sub _remove_defaultports {
	# Now remove the preconfigured root ports
	foreach my $i ( @{$main::config{'ignorerootports'}}) {
		if ( defined $ttys{$i} ) {
			print STDERR "unsetting $i in ttys array\n"
				if ( $main::opts{'d'} );
			delete $ttys{$i};
		}
	}
}

sub check_umask {
	print STDERR "In Security::rootlogin::check_umask()\n" if ( $main::opts{'d'} );

	$umask = umask;
}

sub print {
	print STDERR "In Security::rootlogin::print()\n" if ( $main::opts{'d'} );
	
	&print_header("\n\nRoot has world writable umask value ($umask)\n")
		if ( ! $umask & 0x2 );

	&print_header("\n\nRoot has group writable umask value ($umask)\n")
		if ( ! $umask & 0x20 );

	&print_header("\n\nRoot login allowed on all console ports\n")
		if ( ! $rootlogin );

	&print_data ("\n\nchecking for ports root is allowed to login to:\n", %ttys);
}

1;

__END__

=head1 NAME

Security::rootlogin - root login control module

=head1 SYNOPSIS

Security::rootlogin::init();

Security::rootlogin::print();

=head1 DESCRIPTION

Verifies that root is allowed to login on only certain ports.

=head1 API

This module implements the default API:

=over 4

=item C<init()>

Calls C<read_config (ignorerootports)> to get the list of ports that root
is allowed to login to.

=item C<print()>

Prints out warnings if root's C<umask> is world or group writable.

Calls C<print_data ()> to print out the list of accounts root can login on
that we do not know about.

=back

It also provides the following API calls for the OS specific modules
to use.

=over 4

=item C<check_securettys>

For OSes that have F</etc/securetty> file to control root console ports. 

=item C<check_defaultlogin>

For OSes that have F</etc/default/login> file to control root console ports. 

=item C<check_ttys>

For OSes that have F</etc/ttys> file to control root console ports. 

=item C<check_access>

For OSes that have F</etc/login.access> file to control root console ports.

=back

=head1 CONFIGURATION OPTIONS

The script will call C<read_config> on the following config file options:

=over 4

=item C<ignorerootports>

The list of ports that root is allowed to login on.

=back

=head1 SEE ALSO

L<umask>

=cut

