#
#  Copyright 1996-2014 Jim Pirzyk / pirzyk.org
#  All rights reserved.
#
#  Security::util::groups - Cached getgrent data
#
package Security::util::groups;

use User::grent;
use Exporter;
use strict;
use warnings;
use Security::util::uniq qw(uniq_obj);

our @ISA = qw(Exporter);
our @EXPORT_OK = qw( init check get );

our ($gids, $groups);
#our @METHODS = ( "name", "passwd", "gid", "members" );
our @METHODS = ( "name", "passwd", "gid" );

sub init {
	setgrent;
	while (my $gr = getgrent) {
		# Skip the YP map entry...
		if ( defined $gr ) {
			if ( $gr->name ne '+' ) {

				# Apply patch for bug in perl's getgrent on AIX, wrt gid = -2
				$gr->gid = 4294967294
					if ($gr->gid == -2 &&  $^O eq 'aix');

				# remove duplicates from the list
				# Mac OS X has this problem because getgr* routines are configured
				# to look into files as well as directoryservices,
				# both have the same duplicate data.
				push @{$groups->{$gr->name}}, $gr
					if ( $^O ne 'darwin' || &Security::util::uniq::uniq_obj($gr, $groups->{$gr->name}));
				push @{$gids->{$gr->gid}}, $gr
					if ( $^O ne 'darwin' || &Security::util::uniq::uniq_obj($gr, $gids->{$gr->gid}));

			# If we are not running YP, then flag the '+' as an error
			} elsif ( ! exists $main::config{'yp'} || ! defined $main::config{'yp'} || ! length $main::config{'yp'} ) {
				$Security::group::badlogin{$gr->name} = 1;
			}
		}
	}
	endgrent;
}

sub _bygid {
	if ( ref $groups->{$a} eq 'ARRAY' && ref $groups->{$b} eq 'ARRAY' ) {
		$groups->{$a}->[0]->gid <=> $groups->{$b}->[0]->gid;
	} else {
		return 1 if ( ref $groups->{$a} eq 'ARRAY' );
		return -1 if ( ref $groups->{$b} eq 'ARRAY' );
	}
}

sub get {
	my ($field, $value, $subfield) = @_;
	my (@res, $i);

	if ( defined $value && defined $subfield ) {
		if ( $field eq 'gid' && exists $gids->{$value} ) {
			if ( scalar @{$gids->{$value}} > 1 ) {
				for ( $i = 0; $i < scalar @{$gids->{$value}}; $i++ ) {
					push @res, $gids->{$value}->[$i]->$subfield;
				}
				return @res;
			} else {
				return $gids->{$value}->[0]->$subfield;
			}
		} elsif ( exists $groups->{$value} ) {
			if ( scalar @{$groups->{$value}} > 1 ) {
				for ( $i = 0; $i < scalar @{$groups->{$value}}; $i++ ) {
					push @res, $groups->{$value}->[$i]->$subfield;
				}
				return @res;
			} else {
				return $groups->{$value}->[0]->$subfield;
			}
		} else {
			return undef;
		}
	} else {
		return $field eq 'gid' 
			? sort keys %{$gids}
			: sort _bygid keys %{$groups};
	}
}

sub check {
	my ($field, $value) = @_;
	my ($res);

	if ( $field eq 'gid' ) {
		$res = exists $gids->{$value};
	} else {
		$res = exists $groups->{$value};
	}

	return $res;
}

1;

__END__

=head1 NAME

Security::util::groups - module for checking against the group file.

=head1 SYNOPSIS

Security::util::groups::check($field, $value)

=head1 DESCRIPTION

A wrapper routine around C<User::grent::getgrent> calls to return the
group file information.

=head1 API

The API calls provided by this module

=over 4

=item C<init ()>

Reads in all the group information.

=item C<get ($field [, $value, $subfield])>

If I<$field> is the only argument, C<get()> will return a list all the
values, in gid order.  I<$field> can be either C<name> or C<gid>.  If
I<$value> and I<$subfield> are defined, C<get()> will treat $value as
either a login name or gid and return the specific $subfield value from
the group file.  

=item C<check ($field, $value)>

Returns true if I<$field> has I<$value>.  I<$field> can be either C<name>
or C<gid>.

=back

=head1 CONFIGURATION OPTIONS

None.

=head1 SEE ALSO

L<User::grent>

L<perl>

=cut
