#!/usr/bin/env perl
#
#  Copyright 1996-2014 Jim Pirzyk / pirzyk.org
#  All rights reserved.
#
#  generate_config.pl - script to generate the security.cf file.
#

use Getopt::Std;
use Sys::Filesystem;
use Security::util::uniq;
use strict;
use warnings "all";

our $opts = 'Im:M:P:R:hy';
our %opts;
our $config = {
	checksumfiles	=> [ '/etc/passwd', '/etc/group', '/etc/shadow', '/etc/sudoers', '/usr/local/etc/sudoers', '/etc/ssh/sshd_config', '/etc/ssh/ssh_config', '/etc/ssh_config', '/etc/sshd_config', '/etc/rkhunter.conf', ],
	ignoredevdirs	=> [ '/dev', '/compat/linux/dev', '/dev/fd', '/selinux' ],
	ignorepkgfiles	=> [ '/var/log/gdm', '/var/run/gdm' ],
	ignorefs		=> [ 'nfs', 'autofs', 'iso9660', 'hsfs', 'cd9660', 'devfs', 'procfs' ],
	intruderfiles	=> [ '/.ncftp', '/root/.ncftp', '/usr/lib/libsn.a', '/dev/reset', '/dev/pmcf1', '/dev/pmcf2', '/dev/pmcf3', '/dev/pmcf4', '/tmp/.w0rm', '/tmp/.X11x', '/tmp/.w0rm0r', '/tmp/w0rmishere' ],
	userfiles		=> [
		'.rhosts:0600', '.shosts:0600', '.netrc:0600', '.gnupg:0700', '.gnupg/secring.gpg:0600', '.gnupg/random_seed:0600', '.pgp:0700', '.pgp/secring.pgp:0600', '.ssh:0700', '.ssh/identity:0600', '.ssh/id_dsa:0600', '.ssh/id_rsa:0600',
		'.bashrc', '.bash_profile', '.bash_login', '.bash_logout',
		'.cshrc', '.emacs', '.exrc', '.forward', '.procmailrc',
		'.fvwmrc', '.inputrc', '.klogin', '.kshrc', '.login',
		'.logout', '.nexrc', '.profile', '.screenrc', '.ssh/config',
		'.ssh/authorized_keys', '.ssh/authorized_keys2', '.ssh/environment',
		'.ssh/known_hosts', '.ssh/rc', '.tcshrc', '.twmrc', '.xsession', '.xinitrc',
		'.Xdefaults', '.Xauthority:600', 
	],
	nosudoowner	=> [ 'no' ],
	nopkgduperr	=> [ 'no' ],
	gpgkeys		=> [
		# CentOS Keys
		'025e513b',	# CentOS-3 Key
		'443e1821',	# CentOS-4 key
		'e8562897',	# CentOS-5 Key
		'c105b9de',	# CentOS-6 Key
		'd0ff3d16',	# CentOS-6 Debuginfo Key
		'fe837f6f',	# CentOS-6 Security Key
		'ef1d6db8',	# CentOS-6 Testing Key

		# Fedora Keys
		'1ac70ce6',	# Fedora Pre Extras Release
		'30c9ecf8',	# Fedora Project (Test Software)
		'4f2a6fd2',	# Fedora Project
		'df9b0ae9',	# Fedora (8 and 9 testing)
		'4ebfc273',	# Fedora (10)
		'0b86274e',	# Fedora (10 testing)
		'd22e77f2',	# Fedora (11)
		'e8e40fde',	# Fedora (13)
		'97a1071f',	# Fedora (14)
		'fdb36b03',	# Fedora (14-s390x)
		'069c8460',	# Fedora (15)
		'3ad31d0b',	# Fedora-SPARC (15)
		'a82ba4b7',	# Fedora (16)
		'10d90a9e',	# Fedora Secondary (16)
		'1aca3465',	# Fedora (17)

		# OpenSuSE keys
		'3d25d3d9',	# SuSE Security Team
		'0dfb3188',	# Open Enterprise Server
		'a1912208',	# Novell Provo Build
		'7e2e3b05',	# Novell Provo Build
		'307e3d54',	# SuSE Package Signing Key
		'9c800aca',	# SuSE Package Signing Key
		'3dbdc284',	# openSUSE Project Signing Key
		'56b4177a',	# openSUSE:Factory OBS Project

		# RedHat Keys
		'2fa658e0',	# Red Hat, Inc. (auxiliary key)
		'37017186',	# Red Hat, Inc. (release key)
		'42193e6b',	# Red Hat, Inc. (RHX key)
		'650d5882',	# Red Hat, Inc. (Security Response Team)
		'897da07a',	# Red Hat, Inc. (Beta Test Software)
		'db42a60e',	# Red Hat, Inc
		'f21541eb',	# Red Hat, Inc. (beta key 2)
		'fd431d51',	# Red Hat, Inc. (release key 2)

		# Debian/Ubuntu Keys
		'437D05B5',	# Ubuntu Archive Automatic Signing Key
		'FBB75451', 	# Ubuntu CD Image Automatic Signing Key
		'C0B21F32',	# Ubuntu Archive Automatic Signing Key (2012)
		'EFE21092',	# Ubuntu CD Image Automatic Signing Key (2012)

		# Extra well known package signing keys
		'217521f6',	# Fedora EPEL
		'0608b895',	# EPEL (6)
		'6b8d79e6',	# Dag Wieers (rpmforge)
		'1aa78495',	# Dries Verachtert
		'4bd6ec30',	# Puppet Labs Release Key
	],
};

sub usage {
	print "Usage: $0 [-hxI] [-P PREFIX] [-m mailhost] [-M MODULE DIR]\n";
	print "\t-h\tThis help screen\n";
	print "\t-m mailhost\tSMTP mail server (defaults to localhost)\n";
	print "\t-M Module DIR\tWhere the Perl modules are installed\n";
	print "\t-P PREFIX DIR\tWhere the program starts\n";
	print "\t-R ROOT DIR\tDirectory where / will be relocated to (for package building)\n";
	print "\t-I\tNon Interactive install\n";
	print "\t-y\tDefault answer is yes (instead of no)\n";
	exit 0;
}

sub ask_question {
	my ($question, $default) = @_;
	my ($result);

	if ( ! exists $opts{'I'} ) {
		printf STDERR "$question [$default] ";
		$result = <STDIN>;
		chomp $result;
	}

	$result = $default if ( ! defined $result || ! length $result );

	return $result;
}

sub write_conf_tag {
	my ($tag, @a) = @_;
	my ($str) = '';

	if ( ! scalar @a ) {
		$str = $tag . ' ' . join (' ', sort(@{$config->{$tag}})) . "\n"
			if ( exists $config->{$tag} && scalar @{$config->{$tag}} );
	} else {
		$str = $tag . ' ' . join (' ', sort(@a)) . "\n";
	}

	return $str;
}

sub generate_config_file {
	my ($contents, @array, $FP);

	# Look through all the mount points for special type of filesystems to exclude from our find call.
     my $fs = new Sys::Filesystem;
	foreach my $fs ( $fs->special_filesystems() ) {
     	push @array, $fs if ( -d $fs );
	}
	$contents = write_conf_tag('ignoredir', @array);
	$contents .= write_conf_tag('ignorefs');

	@array = ();
	foreach my $d (@{$config->{'ignorepkgfiles'}}) {
		push @array, $d if ( -f $d );
	}
	$contents .= write_conf_tag('ignorepkgfiles', @array);
	$contents .= write_conf_tag('nopkgduperr');

	# Check to see what gpg keys we already have installed
	# from the list of known distro keys.
	if ( $^O eq 'linux' ) {
		@array = ();
		if ( -f '/bin/rpm' ) {
			foreach my $d (@{$config->{'gpgkeys'}}) {
				`/bin/rpm -q --quiet gpg-pubkey-$d`;
				push @array, $d if ( $? == 0 );
			}
		} elsif ( -f '/usr/bin/apt-key' ) {
			if (open ($FP, '/usr/bin/apt-key list |' )) {
				while (<$FP>) {
					chomp;
					if ( /^pub / ) {
						my ($pub, $keysize, $key, $date) = split (m|[/\s\t]+|, $_);
						push @array, $key
							if ( grep (/^\Q${key}\E$/, @{$config->{'gpgkeys'}}));
					}
				}
				close $FP;
			}
		}
		$contents .= write_conf_tag('gpgkeys', @array);
	}

	# now look for directories that should containt block and character special devices.
	@array = ();
	foreach my $d (@{$config->{'ignoredevdirs'}}) {
		push @array, $d if ( -d $d );
	}
	# TODO: Need to see if anonftp or some other chroot environment is on the
	# system to add that /dev location.
	$contents .= write_conf_tag('ignoredevdir', @array);

	# Go through all the system accounts and record some interesting info
	use User::pwent;
	my (@nopass, @uid0, @sharedhome, %homedirs, $res);
	while (my $pw = getpwent) {
		
		# If we found an account w/o a password, ask if we should put
		# it in the list of accounts to allow to have no password
		# example of this is the 'sync' account on SunOS 4.x
		if (! defined $pw->passwd || ! length ($pw->passwd)) {
			$res = ask_question ('Should this account have no password (' . $pw->name . ')?', ( exists $opts{'y'} ? 'yes': 'no'));
			push @nopass, $pw->name
				if ( $res eq 'yes');
		}

		# Check out alternate root accounts.
		if ( $pw->uid == 0 && $pw->name ne 'root' ) {
			$res = ask_question ('Should this uid 0 account exist (' . $pw->name . ')?', ( exists $opts{'y'} ? 'yes': 'no'));
			push @uid0, $pw->name
				if ( $res eq 'yes');
		}

		# accounts that share home directories
		if ( exists $homedirs{$pw->dir} ) {
			$res = ask_question ('Should ' . $pw->name . ' and ' . $homedirs{$pw->dir} . ' share home directories?', ( exists $opts{'y'} ? 'yes': 'no'));
			push @sharedhome, ( $pw->name, $homedirs{$pw->dir} )
				if ( $res eq 'yes');
		}
		$homedirs{$pw->dir} = $pw->name;
	}
	$contents .= write_conf_tag('ignoreuid', @uid0);
	$contents .= write_conf_tag('ignorepass', @nopass);
	$contents .= write_conf_tag('nosudoowner');
	$contents .= write_conf_tag('ignoreusers', &Security::util::uniq::uniq(@sharedhome));
	$contents .= write_conf_tag('userfiles');
	$contents .= write_conf_tag('intruderfiles');
	
	# Get the SMTP mail server
	my ($mailhost) = ask_question ('SMTP Mail server?', $opts{'m'});
	$contents .= "mailhost $mailhost\n";

	# Now based on the OS, setup the ports root is allowed to login to.
	@array = ();

	# This will not work if we are running as non root, since on linux
	# the /etc/securetty file is read only by root.
	if ( ! $< && $^O eq 'linux' && -f '/etc/securetty' ) {
		if (open ($FP, '/etc/securetty' )) {
			while (<$FP>) {
				chomp;
				s/#.*$//;
				push @array, $_ if ( length $_ );
			}
			close $FP;
			$contents .= write_conf_tag('ignorerootports', @array);
		} else {
			die "Could not open securetty\n";
		}
	} else {
		warn "Could not read securetty because we are not root\n"
			if ( $< && -f '/etc/securetty' );
		$contents .= "ignorerootports console\n";
	}

	# Any other 'special' system files that we should checksum by default
	@array = ();
	foreach my $f (@{$config->{'checksumfiles'}}) {
		push @array, $f
			if ( -e $f );
	}
	# Add our script, the config file and our modules to be checksummed.
	push @array, $opts{'P'} . '/sbin/security',
		$opts{'P'} . '/etc/security/security.cf',
		glob( $opts{'R'} . $opts{'M'} . '/Security/*.pm'),
		glob( $opts{'R'} . $opts{'M'} . '/Security/OS/*.pm'),
		glob( $opts{'R'} . $opts{'M'} . '/Security/util/*.pm');
	# Purge out the ROOTDIR value so it is not in the config file.
	my $i =0;
	map { $array[$i++] =~ s,$opts{'R'},,; } @array
		if ( length $opts{'R'} );
	$contents .= write_conf_tag('checksumfiles', @array);

	# Now record the modules to load 
	# TODO: Need to verify that order is correct.
	my @load = glob( $opts{'R'} . $opts{'M'} . '/Security/*.pm');
	if ( scalar @load ) {
		use File::Basename;
		for ( my $i = 0; $i < scalar @load; $i++ ) {
			$load[$i] = basename $load[$i], '.pm';
		}
		$contents .= 'load ' . join (' ', @load) . "\n";
	} else {
		die "Could not find perl modules\n";
	}

	return $contents;
}

# Start of the main program
&getopts ($opts, \%opts);
usage() if ( $opts{'h'} );
$opts{'P'} = '' if ( ! exists $opts{'P'} || ! defined $opts{'P'} );
$opts{'R'} = '' if ( ! exists $opts{'R'} || ! defined $opts{'R'} );
$opts{'m'} = 'localhost' if ( ! exists $opts{'m'} );

# Now start generating the configuration files...
print &generate_config_file();

exit 0;
